"""The photometric error model."""
from typing import Any, Tuple, Union

import numpy as np
import pandas as pd

from photerr.params import ErrorParams


class PhotometricErrorModel:
    """Base error model from Ivezic 2019.

    References
    ----------
    Ivezic 2019 - https://arxiv.org/abs/0805.2366
    van den Busch 2020 - http://arxiv.org/abs/2007.01846
    Kuijken 2019 - https://arxiv.org/abs/1902.11265
    """

    def __init__(self, *args: ErrorParams, **kwargs: Any) -> None:
        """Create an error model using the passed ErrorParams or keyword overrides.

        If you provide an ErrorParams object, those error parameters are used.
        If you provide keyword arguments, those values are used to override the defaults
        defined in the ErrorParams class.
        If you pass an ErrorParams object *and* keyword arguments, then the keyword
        values override the values in the passed ErrorParams object instead.
        """
        # check that non-keyword argument is just an ErrorParams object
        args_error_msg = (
            "The only non-keyword argument that can be passed "
            "is a single ErrorParams object."
        )
        if len(args) > 1:
            raise ValueError(args_error_msg)
        elif len(args) == 1 and not isinstance(args[0], ErrorParams):
            raise TypeError(args_error_msg)

        # assemble and save the final ErrorParams
        if len(args) == 0 and len(kwargs) == 0:
            raise ValueError(
                "Please provide either an ErrorParams object or the required keyword "
                "arguments from ErrorParams."
            )
        elif len(args) > 0 and len(kwargs) == 0:
            self._params = args[0]
        elif len(args) == 0 and len(kwargs) > 0:
            self._params = ErrorParams(**kwargs)
        else:
            self._params = args[0].copy()
            self._params.update(**kwargs)

        # make a list of the bands
        self._bands = [band for band in self._params.nVisYr.keys()]

        # calculate all of the 5-sigma limiting magnitudes
        self._calculate_m5()

        # calculate the limits for sigLim
        self._sigLim = self.getLimitingMags(self.params.sigLim, coadded=True)

    @property
    def params(self) -> ErrorParams:
        """The error model parameters in an ErrorParams objet."""  # noqa: D401
        return self._params

    def _calculate_m5(self) -> None:
        """Calculate the single-visit 5-sigma limiting magnitudes.

        Uses Eq. 6 from Ivezic 2019.
        However, if m5 has been explicitly passed for any bands, those values are
        used instead.
        """
        # calculate the m5 limiting magnitudes using Eq. 6
        m5 = {
            band: self.params.Cm[band]
            + 0.50 * (self.params.msky[band] - 21)
            + 2.5 * np.log10(0.7 / self.params.theta[band])
            + 1.25 * np.log10(self.params.tvis / 30)
            - self.params.km[band] * (self.params.airmass - 1)
            for band in self.params.Cm
        }
        m5.update(self.params.m5)

        self._all_m5 = {band: m5[band] for band in self._bands}

    def _get_area_ratio_auto(
        self, majors: np.ndarray, minors: np.ndarray, bands: list
    ) -> np.ndarray:
        """Get the ratio between PSF area and galaxy aperture area for "auto" model.

        Parameters
        ----------
        majors : np.ndarray
            The semi-major axes of the galaxies in arcseconds
        minors : np.ndarray
            The semi-minor axes of the galaxies in arcseconds
        bands : list
            The list of bands to calculate ratios for

        Returns
        -------
        np.ndarray
            The ratio of aperture size to PSF size for each band and galaxy.
        """
        # get the psf size for each band
        psf_size = np.array([self.params.theta[band] for band in bands])

        # convert PSF FWHM to Gaussian sigma
        psf_sig = psf_size / 2.355

        # calculate the area of the psf in each band
        A_psf = np.pi * psf_sig**2

        # calculate the area of the galaxy aperture in each band
        a_ap = np.sqrt(psf_sig[None, :] ** 2 + (2.5 * majors[:, None]) ** 2)
        b_ap = np.sqrt(psf_sig[None, :] ** 2 + (2.5 * minors[:, None]) ** 2)
        A_ap = np.pi * a_ap * b_ap

        # return their ratio
        return A_ap / A_psf

    def _get_area_ratio_gaap(
        self, majors: np.ndarray, minors: np.ndarray, bands: list
    ) -> np.ndarray:
        """Get the ratio between PSF area and galaxy aperture area for "gaap" model.

        Parameters
        ----------
        majors : np.ndarray
            The semi-major axes of the galaxies in arcseconds
        minors : np.ndarray
            The semi-minor axes of the galaxies in arcseconds
        bands : list
            The list of bands to calculate ratios for

        Returns
        -------
        np.ndarray
            The ratio of aperture size to PSF size for each band and galaxy.
        """
        # get the psf size for each band
        psf_size = np.array([self.params.theta[band] for band in bands])

        # convert PSF FWHM to Gaussian sigma
        psf_sig = psf_size / 2.355

        # convert min/max aperture diameter to Gaussian sigma
        aMin_sig = self.params.aMin / 2.355
        aMax_sig = self.params.aMax / 2.355

        # convert galaxy half-light radii to Gaussian sigmas
        # this conversion factor comes from half-IQR -> sigma
        majors /= 0.6745
        minors /= 0.6745

        # calculate the area of the psf in each band
        A_psf = np.pi * psf_sig**2

        # calculate the area of the galaxy aperture in each band
        a_ap = np.sqrt(psf_sig[None, :] ** 2 + majors[:, None] ** 2 + aMin_sig**2)
        a_ap[a_ap > aMax_sig] = aMax_sig
        b_ap = np.sqrt(psf_sig[None, :] ** 2 + minors[:, None] ** 2 + aMin_sig**2)
        b_ap[b_ap > aMax_sig] = aMax_sig
        A_ap = np.pi * a_ap * b_ap

        # return their ratio
        return A_ap / A_psf

    def _get_NSR(
        self, mags: np.ndarray, majors: np.ndarray, minors: np.ndarray, bands: list
    ) -> np.ndarray:
        """Calculate the noise-to-signal ratio.

        Uses Eqs 4 and 5 from Ivezic 2019.
        If using extended errors, also rescales with square root of area ratio.

        Parameters
        ----------
        mags : np.ndarray
            The magnitudes of the galaxies
        majors : np.ndarray
            The semi-major axes of the galaxies in arcseconds
        minors : np.ndarray
            The semi-minor axes of the galaxies in arcseconds
        bands : list
            The list of bands the galaxy is observed in

        Returns
        -------
        np.ndarray
            The ratio of aperture size to PSF size for each band and galaxy.
        """
        # get the 5-sigma limiting magnitudes for these bands
        m5 = np.array([self._all_m5[band] for band in bands])
        # and the values for gamma
        gamma = np.array([self.params.gamma[band] for band in bands])
        # and the number of visits per year
        nVisYr = np.array([self.params.nVisYr[band] for band in bands])

        # calculate x as defined in the paper
        x = 10 ** (0.4 * (mags - m5))

        # calculate the NSR for a single visit
        with np.errstate(invalid="ignore"):
            nsrRandSingleExp = np.sqrt((0.04 - gamma) * x + gamma * x**2)

        # calculate the NSR for the stacked image
        nStackedObs = nVisYr * self.params.nYrObs
        nsrRand = nsrRandSingleExp / np.sqrt(nStackedObs)

        # rescale according to the area ratio
        if self.params.extendedType == "auto":
            A_ratio = self._get_area_ratio_auto(majors, minors, bands)
        elif self.params.extendedType == "gaap":
            A_ratio = self._get_area_ratio_gaap(majors, minors, bands)
        else:
            A_ratio = 1  # type: ignore
        nsrRand *= np.sqrt(A_ratio)

        # get the irreducible system NSR
        if self.params.highSNR:
            nsrSys = self.params.sigmaSys
        else:
            nsrSys = 10 ** (self.params.sigmaSys / 2.5) - 1

        # calculate the total NSR
        nsr = np.sqrt(nsrRand**2 + nsrSys**2)

        return nsr

    def _get_obs_and_errs(
        self,
        mags: np.ndarray,
        majors: np.ndarray,
        minors: np.ndarray,
        bands: list,
        rng: np.random.Generator,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the noise-to-signal ratio.

        Uses Eqs 4 and 5 from Ivezic 2019.
        If using extended errors, also rescales with square root of area ratio.

        Parameters
        ----------
        mags : np.ndarray
            The magnitudes of the galaxies
        majors : np.ndarray
            The semi-major axes of the galaxies in arcseconds
        minors : np.ndarray
            The semi-minor axes of the galaxies in arcseconds
        bands : list
            The list of bands the galaxy is observed in
        rng : np.random.Generator
            A numpy random number generator

        Returns
        -------
        np.ndarray
            The ratio of aperture size to PSF size for each band and galaxy.
        """
        # get the NSR for all galaxies
        nsr = self._get_NSR(mags, majors, minors, bands)

        if self.params.highSNR:
            # in the high SNR approximation, mag err ~ nsr, and we can model
            # errors as Gaussian in magnitude space

            # calculate observed magnitudes
            obsMags = rng.normal(loc=mags, scale=nsr)

            # if ndMode == sigLim, then clip all magnitudes at the n-sigma limit
            if self.params.ndMode == "sigLim":
                sigLim = np.array([self._sigLim[band] for band in bands])
                obsMags = np.clip(obsMags, None, sigLim)

            # if decorrelate, then we calculate new errors using the observed mags
            if self.params.decorrelate:
                nsr = self._get_NSR(obsMags, majors, minors, bands)

            obsMagErrs = nsr

        else:
            # in the more accurate error model, we acknowledge err != nsr,
            # and we model errors as Gaussian in flux space

            # calculate observed magnitudes
            fluxes = 10 ** (mags / -2.5)
            obsFluxes = fluxes * (1 + rng.normal(scale=nsr))
            if self.params.ndMode == "abs":
                obsFluxes = np.abs(obsFluxes)
            with np.errstate(divide="ignore"):
                obsMags = -2.5 * np.log10(np.clip(obsFluxes, 0, None))

            # if ndMode == sigLim, then clip all magnitudes at the n-sigma limit
            if self.params.ndMode == "sigLim":
                sigLim = np.array([self._sigLim[band] for band in bands])
                obsMags = np.clip(obsMags, None, sigLim)

            # if decorrelate, then we calculate new errors using the observed mags
            if self.params.decorrelate:
                nsr = self._get_NSR(obsMags, majors, minors, bands)

            obsMagErrs = 2.5 * np.log10(1 + nsr)

        return obsMags, obsMagErrs

    def __call__(
        self, catalog: pd.DataFrame, random_state: Union[np.random.Generator, int]
    ) -> pd.DataFrame:
        """Calculate photometric errors for the catalog and return in DataFrame.

        Parameters
        ----------
        catalog : pd.DataFrame
            The input catalog of galaxies in a pandas DataFrame.
        random_sate : np.random.Generator or int
            The random state. Can either be a numpy random generator
            (e.g. np.random.default_rng(42)), or an integer, which is then used
            to seed np.random.default_rng.

        Returns
        -------
        pd.DataFrame
        """
        # set the rng
        if isinstance(random_state, np.random.Generator):
            rng = random_state
        elif isinstance(random_state, int):
            rng = np.random.default_rng(random_state)
        else:
            raise TypeError("random_state must be a numpy random generator or an int.")

        # get the bands we will calculate errors for
        bands = [band for band in catalog.columns if band in self._bands]

        # get the numpy array of magnitudes
        mags = catalog[bands].to_numpy()

        # get the semi-major and semi-minor axes
        if self.params.extendedType == "auto" or self.params.extendedType == "gaap":
            majors = catalog[self.params.majorCol].to_numpy()
            minors = catalog[self.params.minorCol].to_numpy()
        else:
            majors = None
            minors = None

        # get observed magnitudes and errors
        obsMags, obsMagErrs = self._get_obs_and_errs(mags, majors, minors, bands, rng)

        # flag all non-finite values with the ndFlag
        if self.params.ndMode == "flag":
            obsMags[~np.isfinite(obsMags)] = self.params.ndFlag
            obsMagErrs[~np.isfinite(obsMags)] = self.params.ndFlag

        # save the observations in a DataFrame
        errDf = pd.DataFrame(obsMagErrs, columns=[f"{band}_err" for band in bands])
        if self.params.errLoc == "alone":
            obsCatalog = errDf
        else:
            magDf = catalog.copy()
            magDf[bands] = obsMags
            obsCatalog = pd.concat([magDf, errDf], axis=1)

        if self.params.errLoc == "after":
            # reorder the columns so that the error columns come right after the
            # respective magnitude columns
            columns = catalog.columns.tolist()
            for band in bands:
                columns.insert(columns.index(band) + 1, f"{band}_err")
            obsCatalog = obsCatalog[columns]

        return obsCatalog

    def getLimitingMags(self, nSigma: float = 5, coadded: bool = True) -> dict:
        """Return the limiting magnitudes for point sources in all bands.

        This method essentially reverse engineers the _get_NSR method so that we
        calculate what magnitude results in NSR = 1/nSigma.
        (NSR is noise-to-signal ratio; NSR = 1/SNR)

        Parameters
        ----------
        nSigma : float; default=5
            Sets which limiting magnitude to return. E.g. nSigma=1 returns the 1-sigma
            limiting magnitude. In other words, nSigma is equal to the signal-to-noise
            ratio (SNR) of the limiting magnitudes.
        coadded : bool; default=True
            If True, returns the limiting magnitudes for a coadded image. If False,
            returns the limiting magnitudes for a single visit.

        Returns
        -------
        dict
            The dictionary of limiting magnitudes
        """
        bands = self._bands

        # if nSigma is zero, return infinite magnitude limits
        if np.isclose(0, nSigma):
            return {band: np.inf for band in bands}

        # get the 5-sigma limiting magnitudes for these bands
        m5 = np.array([self._all_m5[band] for band in bands])
        # and the values for gamma
        gamma = np.array([self.params.gamma[band] for band in bands])
        # and the number of visits per year
        nVisYr = np.array([self.params.nVisYr[band] for band in bands])

        # get the number of exposures
        if coadded:
            nStackedObs = nVisYr * self.params.nYrObs
        else:
            nStackedObs = 1  # type: ignore

        # get the irreducible system error
        if self.params.highSNR:
            nsrSys = self.params.sigmaSys
        else:
            nsrSys = 10 ** (self.params.sigmaSys / 2.5) - 1

        # calculate the random NSR that a single exposure must have
        nsrRandSingleExp = np.sqrt((1 / nSigma**2 - nsrSys**2) * nStackedObs)

        # calculate the value of x that corresponds to this NSR
        # this is just the quadratic equation applied to Eq 5 from Ivezic 2019
        x = (
            (gamma - 0.04)
            + np.sqrt((gamma - 0.04) ** 2 + 4 * gamma * nsrRandSingleExp**2)
        ) / (2 * gamma)

        # convert x to a limiting magnitude
        limiting_mags = m5 + 2.5 * np.log10(x)

        # return as a dictionary
        return dict(zip(bands, limiting_mags))

    def __repr__(self) -> str:  # noqa: D105
        return "Photometric error model with parameters:\n\n" + str(self.params)



"""Photometric error model for Roman."""
from dataclasses import dataclass, field
from typing import Any, Dict

from photerr.model import PhotometricErrorModel
from photerr.params import ErrorParams, param_docstring


@dataclass
class RomanErrorParams(ErrorParams):
    """Parameters for the Roman photometric error model.

    gamma and limiting magnitudes taken from page 4 of Graham 2020.
    nYrObs and nVisYr set = 1, assuming that Roman is point-and-stare.
    """

    __doc__ += param_docstring
    __doc__ += "    Graham 2020 - https://arxiv.org/abs/2004.07885"

    nYrObs: float = 1.0
    nVisYr: Dict[str, float] = field(
        default_factory=lambda: {
            "Y": 1.0,
            "J": 1.0,
            "H": 1.0,
            "F": 1.0,
        }
    )
    gamma: Dict[str, float] = field(
        default_factory=lambda: {
            "Y": 0.04,
            "J": 0.04,
            "H": 0.04,
            "F": 0.04,
        }
    )
    m5: Dict[str, float] = field(
        default_factory=lambda: {
            "Y": 26.9,
            "J": 26.95,
            "H": 26.9,
            "F": 26.25,
        }
    )


class RomanErrorModel(PhotometricErrorModel):
    """Photometric error model for Roman."""

    def __init__(self, **kwargs: Any) -> None:
        """Create a Roman error model.

        Keyword arguments override default values in RomanErrorParams.
        """
        super().__init__(RomanErrorParams(), **kwargs)



"""Photometric error model for LSST."""
from dataclasses import dataclass, field
from typing import Any, Dict

from photerr.model import PhotometricErrorModel
from photerr.params import ErrorParams, param_docstring


@dataclass
class LsstErrorParams(ErrorParams):
    """Parameters for the LSST photometric error model.

    Default values taken from pages 11, 12, 26 of Ivezic 2019.
    """

    __doc__ += param_docstring

    tvis: float = 30.0
    nYrObs: float = 10.0
    nVisYr: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 5.6,
            "g": 8.0,
            "r": 18.4,
            "i": 18.4,
            "z": 16.0,
            "y": 16.0,
        }
    )
    gamma: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 0.038,
            "g": 0.039,
            "r": 0.039,
            "i": 0.039,
            "z": 0.039,
            "y": 0.039,
        }
    )
    airmass: float = 1.2
    m5: Dict[str, float] = field(default_factory=lambda: {})
    Cm: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 23.09,
            "g": 24.42,
            "r": 24.44,
            "i": 24.32,
            "z": 24.16,
            "y": 23.73,
        }
    )
    msky: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 22.99,
            "g": 22.26,
            "r": 21.20,
            "i": 20.48,
            "z": 19.60,
            "y": 18.61,
        }
    )
    theta: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 0.81,
            "g": 0.77,
            "r": 0.73,
            "i": 0.71,
            "z": 0.69,
            "y": 0.68,
        }
    )
    km: Dict[str, float] = field(
        default_factory=lambda: {
            "u": 0.491,
            "g": 0.213,
            "r": 0.126,
            "i": 0.096,
            "z": 0.069,
            "y": 0.170,
        }
    )


class LsstErrorModel(PhotometricErrorModel):
    """Photometric error model for Euclid."""

    def __init__(self, **kwargs: Any) -> None:
        """Create an LSST error model.

        Keyword arguments override default values in LsstErrorParams.
        """
        super().__init__(LsstErrorParams(), **kwargs)



"""Parameter objects for the error model."""
from __future__ import annotations

from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, Union

import numpy as np

param_docstring = """

    Parameters
    ----------
    nYrObs : float
        Number of years of observations
    nVisYr : dict
        Mean number of visits per year in each band
    gamma : dict
        A band dependent parameter defined in Ivezic 2019
    m5 : dict
        A dictionary of single visit 5-sigma limiting magnitudes. For any
        bands for which you pass a value in m5, this will be the 5-sigma
        limiting magnitude used, and any values for that band in Cm, msky,
        theta, and km will be ignored.
    tvis : float
        Exposure time in seconds for a single visit
    airmass : float
        The fiducial airmass
    Cm : dict
        A band dependent parameter defined in Ivezic 2019
    msky : dict
        Median zenith sky brightness in each band
    theta : dict
        Median zenith seeing FWHM in arcseconds for each band
    km : dict
        Atmospheric extinction in each band
    sigmaSys : float; default=0.005
        The irreducible error of the system in AB magnitudes. Sets the minimum
        photometric error.
    ndMode : str; default="flag"
        The non-detection mode. I.e. how should the model handle non-detections?
        Non-detections are defined as bands with negative flux.
        There are three options:
            - "flag" - non-detections are flagged using the ndFlag (see below)
            - "sigLim" - magnitudes are clipped at the n-sigma limits. I.e. if
                sigLim=1 below, then all magnitudes greater than the 1-sigma limit
                in each band are replaced with the 1-sigma limiting magnitude.
            - "abs" - the absolute value of negative fluxes are taken before
                converting back to magnitudes. This assures there are no negative
                fluxes so that each galaxy can be assigned a magnitude.
    ndFlag : float; default=np.inf
        Flag for non-detections when ndMode == "flag".
    sigLim : float; default=1
        The n-sigma limit at which to clip magnitudes when ndMode == "sigLim".
    extendedType: str; default="point"
        Whether to use the error model for point sources or extended sources.
        For point sources, use "point". For extended sources, you can use "auto"
        or "gaap". See Notes below for more details on these models.
    aMin : float; default=2.0
        The minimum GAaP aperture diameter in arcseconds.
    aMax : float; default=0.7
        The maximum GAaP aperture diameter in arcseconds.
    majorCol : str; default="major"
        The name of the column containing the semi-major axes of the galaxies (in
        arcseconds). The length scales corresponds to the half-light radius.
    minorCol : str; default="minor"
        The name of the column containing the semi-minor axes of the galaxies (in
        arcseconds). The length scales corresponds to the half-light radius.
    decorrelate : bool; default=True
        Whether or not to decorrelate the photometric errors. If True, after calculating
        observed magnitudes, the errors are re-calculated using the observed magnitudes.
        If False, the original photometric errors are returned. Be warned, however, 
        that in this case, the returned photometric errors are calculated from the true
        magnitudes, and thus provide a deterministic link back to the true magnitudes!
    highSNR : bool; default=False
        Whether to use the high SNR approximations given in Ivezic 2019. If False,
        then Eq. 5 from that paper is used to calculate (N/S)^2 in flux, and errors
        are Gaussian in flux space. If True, Eq. 5 is used to calculate the Gaussian
        variance in magnitude space.
    errLoc: str; default="after"
        Where to place the error columns in the output table. If "after", then the
        error columns will be placed immediately after the corresponding magnitude
        columns. If "end", then all of the error columns will be placed at the end
        of the table. If "alone", then the errors will be returned by themselves.
    renameDict : dict; optional
        A dictionary used to rename the bands in the parameters above that are
        dictionaries. This is useful if you want to use some of the default parameters
        but have given your bands different names. For example, if your bands are named
        "lsst_u", "lsst_g", etc., then you can provide
        renameDict={"u": "lsst_u", "g": "lsst_g", ...}, and all of the default
        parameters will be renamed. Additionally, if you are overriding any of the
        default dictionary parameters, you can provide those overrides using *either*
        the old or the new naming scheme.

    Notes
    -----
    Parameters for the error model from Ivezic 2019. We also implement a more general
    model that is also accurate in the low-SNR regime, and include errors for extended
    sources using the elliptical models from van den Busch 2020 (the "auto" model) and
    Kuijken 2019 (the "gaap" model). The "gaap" model identical to the "auto" model,
    except that it has minimum and maximum aperture sizes.

    When using the model for extended sources, you must provide the semi-major and
    semi-minor axes of the galaxies in arcseconds. These must be in columns whose
    names are given by majorCol and minorCol. In addition, you must provide the PSF
    size for each band in the theta dictionary (see above).

    Whether you are using the point source or extended source models, you must provide
    nVisYr and gamma for every band you wish to calculate errors for. In addition,
    you must provide either:
    - the single-visit 5-sigma limiting magnitude in the m5 dictionary
    - tvis and airmass, plus per-band parameters in the Cm, msky, theta, and km
        dictionaries, which are used to calculate the limiting magnitudes using
        Eq. 6 from Ivezic 2019.

    Note if for any bands you pass a value in the m5 dictionary, the model will use 
    that value for that band, regardless of the values in Cm, msky, theta, and km.

    When instantiating the ErrorParams object, it will determine which bands it has
    enough information to calculate errors for, and throw away all of the extraneous
    parameters. If you expect errors in a certain band, but the error model is not
    calculating errors for that band, check that you have provided all of the required
    parameters for that band.

    References
    ----------
    Ivezic 2019 - https://arxiv.org/abs/0805.2366
    van den Busch 2020 - http://arxiv.org/abs/2007.01846
    Kuijken 2019 - https://arxiv.org/abs/1902.11265
"""


@dataclass
class ErrorParams:
    """Parameters for the photometric error models."""

    __doc__ += param_docstring

    nYrObs: float
    nVisYr: Dict[str, float]
    gamma: Dict[str, float]
    m5: Dict[str, float] = field(default_factory=lambda: {})
    tvis: float = None  # type: ignore
    airmass: float = None  # type: ignore
    Cm: Dict[str, float] = field(default_factory=lambda: {})
    msky: Dict[str, float] = field(default_factory=lambda: {})
    theta: Dict[str, float] = field(default_factory=lambda: {})
    km: Dict[str, float] = field(default_factory=lambda: {})

    sigmaSys: float = 0.005
    sigLim: float = 1
    ndMode: str = "flag"
    ndFlag: float = np.inf

    extendedType: str = "point"
    aMin: float = 0.7
    aMax: float = 2.0
    majorCol: str = "major"
    minorCol: str = "minor"

    decorrelate: bool = True
    highSNR: bool = False
    errLoc: str = "after"

    renameDict: InitVar[Dict[str, str]] = None

    def __post_init__(self, renameDict: Union[Dict[str, str], None]) -> None:
        """Rename bands and remove duplicate parameters."""
        # if renameDict was provided, rename the bands
        if renameDict is not None:
            self.rename_bands(renameDict)

        # clean up the dictionaries
        self._clean_dictionaries()

        # if using extended error types, make sure theta is provided for every band
        if (
            self.extendedType == "auto" or self.extendedType == "gaap"
        ) and self.theta.keys() != self.nVisYr.keys():
            raise ValueError(
                "If using one of the extended error types "
                "(i.e. extendedType == 'auto' or 'gaap'), "
                "then theta must contain an entry for every band."
            )

    def _clean_dictionaries(self) -> None:
        """Remove unnecessary info from all of the dictionaries.

        This means that any bands explicitly listed in m5 will have their corresponding
        info removed from Cm, msky, theta, and km.

        Additionally, we will remove any band from all dictionaries that don't have
        an entry in all of nVisYr, gamma, and (m5 OR Cm + msky + theta + km).
        """
        # keep a running set of all the bands we will calculate errors for
        all_bands = set(self.nVisYr.keys()).intersection(self.gamma.keys())

        # remove the m5 bands from all other parameter dictionaries, and remove
        # bands from all_bands for which we don't have m5 data for
        ignore_dicts = ["m5", "nVisYr", "gamma", "theta"]
        for key, param in self.__dict__.items():
            # get the parameters that are dictionaries
            if isinstance(param, dict) and key not in ignore_dicts:
                # remove bands that we have explicit m5's for
                self.__dict__[key] = {
                    band: val for band, val in param.items() if band not in self.m5
                }

                # update the running list of bands that we have sufficient m5 data for
                all_bands = all_bands.intersection(
                    set(param.keys()).union(set(self.m5.keys()))
                )

        # finally, remove all of the data for bands that we will not be
        # calculating errors for
        for key, param in self.__dict__.items():
            if isinstance(param, dict):
                self.__dict__[key] = {
                    band: val for band, val in param.items() if band in all_bands
                }

        # if there are no bands left, raise an error
        if len(all_bands) == 0:
            raise ValueError(
                "There are no bands left! You probably set one of the dictionary "
                "parameters (i.e. nVisYr, gamma, m5, Cm, msky, theta, or km) such "
                "that there was not enough info to calculate errors for any of the "
                "photometric bands. Remember that for each band, you must have an "
                "entry in nVisYr and gamma, plus either an entry in m5 or an entry "
                "in all of Cm, msky, theta, and km."
            )

    def rename_bands(self, renameDict: Dict[str, str]) -> None:
        """Rename the bands used in the error model.

        This method might be useful if you want to use the default parameters for an
        error model, but have given your bands different names.

        Parameters
        ----------
        renameDict: dict
            A dictionary containing key: value pairs {old_name: new_name}.
            For example map={"u": "lsst_u"} will rename the u band to lsst_u.
        """
        # loop over every parameter
        for key, param in self.__dict__.items():
            # get the parameters that are dictionaries
            if isinstance(param, dict):
                # rename bands in-place
                self.__dict__[key] = {
                    old_name
                    if old_name not in renameDict
                    else renameDict[old_name]: val
                    for old_name, val in param.items()
                }

    def update(self, *args: dict, **kwargs: Any) -> None:
        """Update parameters using either a dictionary or keyword arguments."""
        # if non-keyword arguments passed, make sure it is just a single dictionary
        # and pass it back through as keyword arguments.
        if len(args) > 1:
            raise ValueError(
                "The only non-keyword argument that can be passed is a dictionary."
            )
        elif len(args) == 1:
            if not isinstance(args[0], dict):
                raise TypeError(
                    "The only non-keyword argument that can be passed is a dictionary."
                )
            else:
                self.update(**args[0], **kwargs)

        # update parameters from keywords
        safe_copy = self.copy()
        try:
            for key, param in kwargs.items():
                setattr(self, key, param)

            # call post-init
            self.__post_init__(renameDict=None)
        except ValueError as error:
            self.__dict__ = safe_copy.__dict__
            raise Warning("Aborting update!\n\n" + str(error))

    def copy(self) -> ErrorParams:
        """Return a full copy of this ErrorParams instance."""
        return deepcopy(self)



# flake8: noqa I252, F401
from importlib import metadata

from .euclid import EuclidErrorModel, EuclidErrorParams
from .lsst import LsstErrorModel, LsstErrorParams
from .model import PhotometricErrorModel
from .params import ErrorParams
from .roman import RomanErrorModel, RomanErrorParams

__version__ = metadata.version(__package__)
del metadata
